// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::{
    bundle::{Bundle, BundleType},
    firrtl::{self, ExportOptions},
    intern::Interned,
    module::Module,
    util::streaming_read_utf8::streaming_read_utf8,
};
use clap::{
    builder::{OsStringValueParser, TypedValueParser},
    Parser, Subcommand, ValueEnum, ValueHint,
};
use eyre::{eyre, Report};
use std::{error, ffi::OsString, fmt, io, path::PathBuf, process};
use tempfile::TempDir;

pub type Result<T = (), E = CliError> = std::result::Result<T, E>;

pub struct CliError(Report);

impl fmt::Debug for CliError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.0.fmt(f)
    }
}

impl fmt::Display for CliError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.0.fmt(f)
    }
}

impl error::Error for CliError {}

impl From<io::Error> for CliError {
    fn from(value: io::Error) -> Self {
        CliError(Report::new(value))
    }
}

pub trait RunPhase<Arg> {
    type Output;
    fn run(&self, arg: Arg) -> Result<Self::Output>;
}

#[derive(Parser, Debug, Clone)]
#[non_exhaustive]
pub struct BaseArgs {
    /// the directory to put the generated main output file and associated files in
    #[arg(short, long, value_hint = ValueHint::DirPath, required = true)]
    pub output: Option<PathBuf>,
    /// the stem of the generated main output file, e.g. to get foo.v, pass --file-stem=foo
    #[arg(long)]
    pub file_stem: Option<String>,
    #[arg(long, env = "FAYALITE_KEEP_TEMP_DIR")]
    pub keep_temp_dir: bool,
    #[arg(skip = false)]
    pub redirect_output_for_rust_test: bool,
}

impl BaseArgs {
    fn make_firrtl_file_backend(&self) -> Result<(firrtl::FileBackend, Option<TempDir>)> {
        let (dir_path, temp_dir) = match &self.output {
            Some(output) => (output.clone(), None),
            None => {
                let temp_dir = TempDir::new()?;
                if self.keep_temp_dir {
                    let temp_dir = temp_dir.into_path();
                    println!("created temporary directory: {}", temp_dir.display());
                    (temp_dir, None)
                } else {
                    (temp_dir.path().to_path_buf(), Some(temp_dir))
                }
            }
        };
        Ok((
            firrtl::FileBackend {
                dir_path,
                top_fir_file_stem: self.file_stem.clone(),
                circuit_name: None,
            },
            temp_dir,
        ))
    }
    /// handles possibly redirecting the command's output for Rust tests
    pub fn run_external_command(
        &self,
        mut command: process::Command,
    ) -> io::Result<process::ExitStatus> {
        if self.redirect_output_for_rust_test {
            let (reader, writer) = os_pipe::pipe()?;
            let mut reader = io::BufReader::new(reader);
            command.stderr(writer.try_clone()?);
            command.stdout(writer); // must not leave writer around after spawning child
            command.stdin(process::Stdio::null());
            let mut child = command.spawn()?;
            drop(command); // close writers
            Ok(loop {
                let status = child.try_wait()?;
                streaming_read_utf8(&mut reader, |s| {
                    // use print! so output goes to Rust test output capture
                    print!("{s}");
                    io::Result::Ok(())
                })?;
                if let Some(status) = status {
                    break status;
                }
            })
        } else {
            command.status()
        }
    }
}

#[derive(Parser, Debug, Clone)]
#[non_exhaustive]
pub struct FirrtlArgs {
    #[command(flatten)]
    pub base: BaseArgs,
    #[command(flatten)]
    pub export_options: ExportOptions,
}

#[derive(Debug)]
#[non_exhaustive]
pub struct FirrtlOutput {
    pub file_stem: String,
    pub top_module: String,
    pub output_dir: PathBuf,
    pub temp_dir: Option<TempDir>,
}

impl FirrtlOutput {
    pub fn file_with_ext(&self, ext: &str) -> PathBuf {
        let mut retval = self.output_dir.join(&self.file_stem);
        retval.set_extension(ext);
        retval
    }
    pub fn firrtl_file(&self) -> PathBuf {
        self.file_with_ext("fir")
    }
}

impl FirrtlArgs {
    fn run_impl(&self, top_module: Module<Bundle>) -> Result<FirrtlOutput> {
        let (file_backend, temp_dir) = self.base.make_firrtl_file_backend()?;
        let firrtl::FileBackend {
            top_fir_file_stem,
            circuit_name,
            dir_path,
        } = firrtl::export(file_backend, &top_module, self.export_options)?;
        Ok(FirrtlOutput {
            file_stem: top_fir_file_stem.expect(
                "export is known to set the file stem from the circuit name if not provided",
            ),
            top_module: circuit_name.expect("export is known to set the circuit name"),
            output_dir: dir_path,
            temp_dir,
        })
    }
}

impl<T: BundleType> RunPhase<Module<T>> for FirrtlArgs {
    type Output = FirrtlOutput;
    fn run(&self, top_module: Module<T>) -> Result<Self::Output> {
        self.run_impl(top_module.canonical())
    }
}

impl<T: BundleType> RunPhase<Interned<Module<T>>> for FirrtlArgs {
    type Output = FirrtlOutput;
    fn run(&self, top_module: Interned<Module<T>>) -> Result<Self::Output> {
        self.run(*top_module)
    }
}

/// based on [LLVM Circt's recommended lowering options
/// ](https://circt.llvm.org/docs/VerilogGeneration/#recommended-loweringoptions-by-target)
#[derive(ValueEnum, Copy, Clone, Debug, PartialEq, Eq, Hash)]
#[non_exhaustive]
pub enum VerilogDialect {
    Questa,
    Spyglass,
    Verilator,
    Vivado,
    Yosys,
}

impl fmt::Display for VerilogDialect {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.write_str(self.as_str())
    }
}

impl VerilogDialect {
    pub fn as_str(self) -> &'static str {
        match self {
            VerilogDialect::Questa => "questa",
            VerilogDialect::Spyglass => "spyglass",
            VerilogDialect::Verilator => "verilator",
            VerilogDialect::Vivado => "vivado",
            VerilogDialect::Yosys => "yosys",
        }
    }
    pub fn firtool_extra_args(self) -> &'static [&'static str] {
        match self {
            VerilogDialect::Questa => &["--lowering-options=emitWireInPorts"],
            VerilogDialect::Spyglass => {
                &["--lowering-options=explicitBitcast,disallowExpressionInliningInPorts"]
            }
            VerilogDialect::Verilator => &[
                "--lowering-options=locationInfoStyle=wrapInAtSquareBracket,disallowLocalVariables",
            ],
            VerilogDialect::Vivado => &["--lowering-options=mitigateVivadoArrayIndexConstPropBug"],
            VerilogDialect::Yosys => {
                &["--lowering-options=disallowLocalVariables,disallowPackedArrays"]
            }
        }
    }
}

#[derive(Parser, Debug, Clone)]
#[non_exhaustive]
pub struct VerilogArgs {
    #[command(flatten)]
    pub firrtl: FirrtlArgs,
    #[arg(
        long,
        default_value = "firtool",
        env = "FIRTOOL",
        value_hint = ValueHint::CommandName,
        value_parser = OsStringValueParser::new().try_map(which::which)
    )]
    pub firtool: PathBuf,
    #[arg(long)]
    pub firtool_extra_args: Vec<OsString>,
    /// adapt the generated Verilog for a particular toolchain
    #[arg(long)]
    pub verilog_dialect: Option<VerilogDialect>,
    #[arg(long, short = 'g')]
    pub debug: bool,
}

#[derive(Debug)]
#[non_exhaustive]
pub struct VerilogOutput {
    pub firrtl: FirrtlOutput,
}

impl VerilogOutput {
    pub fn verilog_file(&self) -> PathBuf {
        self.firrtl.file_with_ext("v")
    }
}

impl VerilogArgs {
    fn run_impl(&self, firrtl_output: FirrtlOutput) -> Result<VerilogOutput> {
        let Self {
            firrtl,
            firtool,
            firtool_extra_args,
            verilog_dialect,
            debug,
        } = self;
        let output = VerilogOutput {
            firrtl: firrtl_output,
        };
        let mut cmd = process::Command::new(firtool);
        cmd.arg(output.firrtl.firrtl_file());
        cmd.arg("-o");
        cmd.arg(output.verilog_file());
        if *debug {
            cmd.arg("-g");
            cmd.arg("--preserve-values=named");
        }
        if let Some(dialect) = verilog_dialect {
            cmd.args(dialect.firtool_extra_args());
        }
        cmd.args(firtool_extra_args);
        cmd.current_dir(&output.firrtl.output_dir);
        let status = firrtl.base.run_external_command(cmd)?;
        if status.success() {
            Ok(output)
        } else {
            Err(CliError(eyre!(
                "running {} failed: {status}",
                self.firtool.display()
            )))
        }
    }
}

impl<Arg> RunPhase<Arg> for VerilogArgs
where
    FirrtlArgs: RunPhase<Arg, Output = FirrtlOutput>,
{
    type Output = VerilogOutput;
    fn run(&self, arg: Arg) -> Result<Self::Output> {
        let firrtl_output = self.firrtl.run(arg)?;
        self.run_impl(firrtl_output)
    }
}

#[derive(ValueEnum, Copy, Clone, Debug, PartialEq, Eq, Hash, Default)]
#[non_exhaustive]
pub enum FormalMode {
    #[default]
    BMC,
    Prove,
    Live,
    Cover,
}

impl FormalMode {
    pub fn as_str(self) -> &'static str {
        match self {
            FormalMode::BMC => "bmc",
            FormalMode::Prove => "prove",
            FormalMode::Live => "live",
            FormalMode::Cover => "cover",
        }
    }
}

impl fmt::Display for FormalMode {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.write_str(self.as_str())
    }
}

#[derive(Clone)]
struct FormalAdjustArgs;

impl clap::FromArgMatches for FormalAdjustArgs {
    fn from_arg_matches(_matches: &clap::ArgMatches) -> Result<Self, clap::Error> {
        Ok(Self)
    }

    fn update_from_arg_matches(&mut self, _matches: &clap::ArgMatches) -> Result<(), clap::Error> {
        Ok(())
    }
}

impl clap::Args for FormalAdjustArgs {
    fn augment_args(cmd: clap::Command) -> clap::Command {
        cmd.mut_arg("output", |arg| arg.required(false))
            .mut_arg("verilog_dialect", |arg| {
                arg.default_value(VerilogDialect::Yosys.to_string())
                    .hide(true)
            })
    }

    fn augment_args_for_update(cmd: clap::Command) -> clap::Command {
        Self::augment_args(cmd)
    }
}

#[derive(Parser, Clone)]
#[non_exhaustive]
pub struct FormalArgs {
    #[command(flatten)]
    pub verilog: VerilogArgs,
    #[arg(
        long,
        default_value = "sby",
        env = "SBY",
        value_hint = ValueHint::CommandName,
        value_parser = OsStringValueParser::new().try_map(which::which)
    )]
    pub sby: PathBuf,
    #[arg(long)]
    pub sby_extra_args: Vec<OsString>,
    #[arg(long, default_value_t)]
    pub mode: FormalMode,
    #[arg(long, default_value_t = Self::DEFAULT_DEPTH)]
    pub depth: u64,
    #[arg(long)]
    pub solver: Option<String>,
    #[arg(long)]
    pub smtbmc_extra_args: Vec<String>,
    #[command(flatten)]
    _formal_adjust_args: FormalAdjustArgs,
}

impl fmt::Debug for FormalArgs {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            verilog,
            sby,
            sby_extra_args,
            mode,
            depth,
            solver,
            smtbmc_extra_args,
            _formal_adjust_args: _,
        } = self;
        f.debug_struct("FormalArgs")
            .field("verilog", &verilog)
            .field("sby", &sby)
            .field("sby_extra_args", &sby_extra_args)
            .field("mode", &mode)
            .field("depth", &depth)
            .field("solver", &solver)
            .field("smtbmc_extra_args", &smtbmc_extra_args)
            .finish_non_exhaustive()
    }
}

impl FormalArgs {
    pub const DEFAULT_DEPTH: u64 = 20;
}

#[derive(Debug)]
#[non_exhaustive]
pub struct FormalOutput {
    pub verilog: VerilogOutput,
}

impl FormalOutput {
    pub fn sby_file(&self) -> PathBuf {
        self.verilog.firrtl.file_with_ext("sby")
    }
}

impl FormalArgs {
    fn sby_contents(&self, output: &FormalOutput) -> String {
        let Self {
            verilog: _,
            sby: _,
            sby_extra_args: _,
            mode,
            depth,
            smtbmc_extra_args,
            solver,
            _formal_adjust_args: _,
        } = self;
        struct OptArg<T>(Option<T>);
        impl<T: fmt::Display> fmt::Display for OptArg<T> {
            fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
                if let Some(v) = &self.0 {
                    f.write_str(" ")?;
                    v.fmt(f)
                } else {
                    Ok(())
                }
            }
        }
        let space_solver = OptArg(solver.as_ref());
        let smtbmc_options = smtbmc_extra_args.join(" ");
        let verilog_file = output
            .verilog
            .verilog_file()
            .into_os_string()
            .into_string()
            .ok()
            .expect("verilog file path is not UTF-8");
        let top_module = &output.verilog.firrtl.top_module;
        format!(
            "[options]\n\
            mode {mode}\n\
            depth {depth}\n\
            wait on\n\
            \n\
            [engines]\n\
            smtbmc{space_solver} -- -- {smtbmc_options}\n\
            \n\
            [script]\n\
            read_verilog -sv -formal {verilog_file}\n\
            prep -top {top_module}\n
            "
        )
    }
    fn run_impl(&self, verilog_output: VerilogOutput) -> Result<FormalOutput> {
        let output = FormalOutput {
            verilog: verilog_output,
        };
        let sby_file = output.sby_file();
        std::fs::write(&sby_file, self.sby_contents(&output))?;
        let mut cmd = process::Command::new(&self.sby);
        cmd.arg("-f");
        cmd.arg(sby_file);
        cmd.args(&self.sby_extra_args);
        cmd.current_dir(&output.verilog.firrtl.output_dir);
        let status = self.verilog.firrtl.base.run_external_command(cmd)?;
        if status.success() {
            Ok(output)
        } else {
            Err(CliError(eyre!(
                "running {} failed: {status}",
                self.sby.display()
            )))
        }
    }
}

impl<Arg> RunPhase<Arg> for FormalArgs
where
    VerilogArgs: RunPhase<Arg, Output = VerilogOutput>,
{
    type Output = FormalOutput;
    fn run(&self, arg: Arg) -> Result<Self::Output> {
        let verilog_output = self.verilog.run(arg)?;
        self.run_impl(verilog_output)
    }
}

#[derive(Subcommand, Debug)]
enum CliCommand {
    /// Generate FIRRTL
    Firrtl(FirrtlArgs),
    /// Generate Verilog
    Verilog(VerilogArgs),
    /// Run a formal proof
    Formal(FormalArgs),
}

/// a simple CLI
///
/// Use like:
///
/// ```no_run
/// # use fayalite::prelude::*;
/// # #[hdl_module]
/// # fn my_module() {}
/// use fayalite::cli;
///
/// fn main() -> cli::Result {
///     cli::Cli::parse().run(my_module())
/// }
/// ```
///
/// You can also use it with a larger [`clap`]-based CLI like so:
///
/// ```no_run
/// # use fayalite::prelude::*;
/// # #[hdl_module]
/// # fn my_module() {}
/// use clap::{Subcommand, Parser};
/// use fayalite::cli;
///
/// #[derive(Subcommand)]
/// pub enum Cmd {
///     #[command(flatten)]
///     Fayalite(cli::Cli),
///     MySpecialCommand {
///         #[arg(long)]
///         foo: bool,
///     },
/// }
///
/// #[derive(Parser)]
/// pub struct Cli {
///     #[command(subcommand)]
///     cmd: Cmd, // or just use cli::Cli directly if you don't need more subcommands
/// }
///
/// fn main() -> cli::Result {
///     match Cli::parse().cmd {
///         Cmd::Fayalite(v) => v.run(my_module())?,
///         Cmd::MySpecialCommand { foo } => println!("special: foo={foo}"),
///     }
///     Ok(())
/// }
/// ```
#[derive(Parser, Debug)]
// clear things that would be crate-specific
#[command(name = "Fayalite Simple CLI", about = None, long_about = None)]
pub struct Cli {
    #[command(subcommand)]
    subcommand: CliCommand,
}

impl clap::Subcommand for Cli {
    fn augment_subcommands(cmd: clap::Command) -> clap::Command {
        CliCommand::augment_subcommands(cmd)
    }

    fn augment_subcommands_for_update(cmd: clap::Command) -> clap::Command {
        CliCommand::augment_subcommands_for_update(cmd)
    }

    fn has_subcommand(name: &str) -> bool {
        CliCommand::has_subcommand(name)
    }
}

impl<T> RunPhase<T> for Cli
where
    FirrtlArgs: RunPhase<T, Output = FirrtlOutput>,
{
    type Output = ();
    fn run(&self, arg: T) -> Result<Self::Output> {
        match &self.subcommand {
            CliCommand::Firrtl(c) => {
                c.run(arg)?;
            }
            CliCommand::Verilog(c) => {
                c.run(arg)?;
            }
            CliCommand::Formal(c) => {
                c.run(arg)?;
            }
        }
        Ok(())
    }
}

impl Cli {
    /// forwards to [`clap::Parser::parse()`] so you don't have to import [`clap::Parser`]
    pub fn parse() -> Self {
        clap::Parser::parse()
    }
    /// forwards to [`RunPhase::run()`] so you don't have to import [`RunPhase`]
    pub fn run<T>(&self, top_module: T) -> Result<()>
    where
        Self: RunPhase<T, Output = ()>,
    {
        RunPhase::run(self, top_module)
    }
}
